import 'dart:ffi';
import 'dart:io' show Platform;

import 'package:ffi/ffi.dart';
import 'dart:typed_data';

import 'dart:convert';

DynamicLibrary openLibrary(String name, String path) {
  name = 'lib' + name;
  return Platform.isAndroid || Platform.isLinux
      ? DynamicLibrary.open(path + name + '.so')
      : Platform.isWindows
          ? DynamicLibrary.open(path + name + '.dll')
          : DynamicLibrary.process();
}

class AbiInternalError {
  static const int ok = 0x0;
  static const int invalidArg = 0x1;
  static const int nullPtr = 0x2;
  static const int abort = 0x3;
  static const int callbackException = 0x4;
  static const int undefinedException = 0x5;
  static const int unimplemented = 0x6;
  static const int type = 0x7;
  static const int notAllowedOperation = 0x8;
  static const int noDefaultConstructor = 0x9;

  static void handleError(int error, String message) {
    switch (error) {
      case AbiInternalError.ok:
        return;
      case AbiInternalError.invalidArg:
        throw ArgumentError(message);
      default:
        throw Exception();
    }
  }
}

class AbiStreamValue extends Struct {
  @Int64()
  int handle;
  Pointer<Void> value;
}

class AbiStream extends Struct {
  @Int64()
  int handle;
}

class AbiVariant extends Struct {
  @Int64()
  int variant;
  Pointer<Void> data;
}

class AbiArray extends Struct {
  @Int64()
  int length;
  Pointer<Void> data;
}

extension AbiArrayEx on AbiArray {
  Pointer<AbiArray> asPtr() {
    final result = allocate<AbiArray>();
    result.ref.data = data;
    result.ref.length = length;
    return result;
  }
}

class AbiMap extends Struct {
  @Int64()
  int length;
  Pointer<Void> key;
  Pointer<Void> data;
}

extension AbiMapEx on AbiMap {
  Pointer<AbiMap> asPtr() {
    final result = allocate<AbiMap>();
    result.ref.data = data;
    result.ref.key = data;
    result.ref.length = length;
    return result;
  }
}

class AbiBytes extends Struct {
  @Int64()
  int length;
  Pointer<Uint8> data;
}

extension BytesPointer on Pointer<AbiBytes> {
  void release() {
    free(ref.data);
    free(this);
  }

  Uint8List asUint8List() {
    return Uint8List.fromList(ref.data.asTypedList(ref.length));
  }
}

extension BytesInto on Uint8List {
  Pointer<AbiBytes> asAbiBytes() {
    final result = allocate<Uint8>(count: length + 0x8);
    final listSource = result.asTypedList(length);
    listSource.setAll(0, this);

    final bytes = allocate<AbiBytes>().ref
      ..length = length
      ..data = result;

    return bytes.addressOf;
  }
}

class AbiString extends Struct {
  @Int64()
  int length;
  Pointer<Uint8> data;
}

extension AbiStringPointer on Pointer<AbiString> {
  String asString() {
    return utf8.decode(Uint8List.view(
      ref.data.asTypedList(ref.length).buffer,
      0,
      ref.length,
    ));
  }

  void release() {
    free(ref.data);
    free(this);
  }
}

extension AbiStringInto on String {
  Pointer<AbiString> asAbiString() {
    final dataString = utf8.encode(this);
    final resultData = allocate<Uint8>(count: dataString.length + 0x8);
    resultData.asTypedList(dataString.length).setAll(0, dataString);

    final utf8String = allocate<AbiString>().ref
      ..data = resultData
      ..length = dataString.length;

    return utf8String.addressOf;
  }
}

extension AbiVariantPointer on Pointer<AbiVariant> {
  void release() {
    free(this);
  }
}

extension PointerUint8Mem on Pointer<Uint8> {
  void release() => free(this);
}

extension PointerInt8Mem on Pointer<Int8> {
  void release() => free(this);
}

extension PointerUint16Mem on Pointer<Uint16> {
  void release() => free(this);
}

extension PointerInt16Mem on Pointer<Int16> {
  void release() => free(this);
}

extension PointerUint32Mem on Pointer<Uint32> {
  void release() => free(this);
}

extension PointerInt32Mem on Pointer<Int32> {
  void release() => free(this);
}

extension PointerUint64Mem on Pointer<Uint64> {
  void release() => free(this);
}

extension PointerInt64Mem on Pointer<Int64> {
  void release() => free(this);
}

extension PointerDoubleMem on Pointer<Double> {
  void release() => free(this);
}

extension PointerFloatMem on Pointer<Float> {
  void release() => free(this);
}
